from clipboard.clipboard import Clipboard
from clipboard.clipboard import ClipboardFormat
from clipboard.clipboard import get_available_formats
from clipboard.clipboard import get_clipboard
from clipboard.clipboard import set_clipboard
from clipboard.errors import ClipboardError
from clipboard.errors import EmptyClipboardError
from clipboard.errors import FormatNotSupportedError
from clipboard.errors import GetClipboardError
from clipboard.errors import GetFormatsError
from clipboard.errors import LockError
from clipboard.errors import OpenClipboardError
from clipboard.errors import SetClipboardError
from clipboard.formats import get_format_name
from clipboard.html_clipboard import HTML_ENCODING


__all__ = [
    "Clipboard",
    "HTML_ENCODING",
    # Formats
    "ClipboardFormat",
    "get_format_name",
    # Convenience Functions
    "get_available_formats",
    "get_clipboard",
    "set_clipboard",
    # Errors
    "ClipboardError",
    "EmptyClipboardError",
    "FormatNotSupportedError",
    "GetClipboardError",
    "GetFormatsError",
    "LockError",
    "OpenClipboardError",
    "SetClipboardError",
]



"""Clipboard Formats"""

import ctypes
from enum import Enum
from enum import EnumMeta
from typing import Any
from typing import Optional

from clipboard._c_interface import CF_HTML
from clipboard._c_interface import CF_RTF
from clipboard._c_interface import GetClipboardFormatNameA


class ExtendedEnum(EnumMeta):
    """Extended Enum Meta Class"""

    def __contains__(cls, item: Any):
        return any(
            [
                item in cls.names,  # type: ignore
                item in cls.values,  # type: ignore
                item in ClipboardFormat.__members__.values(),
            ]
        )


class classproperty:
    def __init__(self, func):
        self.func = func

    def __get__(self, obj, owner):
        return self.func(owner)


class ClipboardFormat(Enum, metaclass=ExtendedEnum):
    """Clipboard Formats"""

    # Constants
    CF_TEXT = 1
    """ANSI text format. Lines end with CR-LF. Ends with null character."""
    CF_UNICODETEXT = 13
    """Unicode text format. Lines end with CR-LF. Ends with a null character."""
    CF_LOCALE = 16
    CF_DIB = 8
    """A memory object containing a
    [BITMAPINFO](https://learn.microsoft.com/en-us/windows/win32/api/wingdi/
    ns-wingdi-bitmapinfo) structure followed by the bitmap bits."""

    # Registered Formats
    CF_HTML = CF_HTML
    CF_RTF = CF_RTF
    HTML_Format = 49418

    # Aliases
    text = CF_UNICODETEXT  # alias
    html = HTML_Format  # alias
    HTML = html  # alias
    rtf = CF_RTF  # alias

    @classproperty
    def values(cls):
        """Get the values of the enum."""
        return [i.value for i in cls]

    @classproperty
    def names(cls):
        """Get the names of the enum."""
        return [i.name for i in cls]

    def __str__(self):
        return f"{str(self.value)} {str(self.name)}"

    def __eq__(self, other):
        if isinstance(self, type(other)):
            return self.name == other.name and self.value == other.value
        elif isinstance(other, int):
            return self.value == other
        else:
            return False


def get_format_name(format_code: int) -> Optional[str]:
    """Get the name of the format by its number.

    C function does not work for standard types (e.g. 1 for CF_TEXT).
    So, this function will use ClipboardFormat for those in the standard.

    Returns
    -------
    str, optional
        The name of the format.
        None if the format is not found.
    """
    # Built-In
    if (
        format_code in ClipboardFormat.values
    ):  # pylint: disable=unsupported-membership-test
        return ClipboardFormat(format_code).name

    buffer_size = 256
    buffer = ctypes.create_string_buffer(buffer_size)
    return_code = GetClipboardFormatNameA(
        format_code,
        buffer,
        buffer_size,
    )

    # Failed
    if return_code == 0:
        last_error: int = ctypes.get_last_error()
        if last_error == 0:
            # No Error
            return None
        if last_error == 87:
            # This indicates that the first parameter is not a valid clipboard format.
            return None
        error = ctypes.WinError(last_error)
        raise error

    # ansi string
    format_name: str = buffer.value.decode("utf-8")

    return format_name



"""Code for handling HTML clipboard data."""

from typing import List
from typing import Optional


ENCODING = "UTF-8"
HTML_ENCODING = ENCODING


class HTMLTemplate:
    """Windows HTML template for storing clipboard HTML data."""

    version = 1.0

    def __init__(self, content: str = "") -> None:
        self.fragments: List[str] = []

        # Padding for byte counts to allow for any reasonable size. e.g. "0000001001"
        self.byte_padding: int = 10

        # Byte Counts
        self.start_html: int = -1
        self.end_html: int = -1

        self.start_fragment: int = -1
        self.end_fragment: int = -1

        # Optional
        self.start_selection: Optional[str] = None
        self.end_selection: Optional[str] = None

        # Target Content
        self.content: str = content

    def generate(self) -> str:
        """Generate the HTML template."""
        fragments: List[str] = (
            self.fragments if self.fragments else [self.content]
        )

        # Generate Fragments
        result: str = self._generate_fragments(fragments)
        # Generate HTML
        result = self._generate_html(result)
        # Add Header
        result = self._generate_header(result)
        # Get Byte Counts
        result = self._update_byte_counts(result)

        return result

    @staticmethod
    def _generate_fragments(fragments: List) -> str:
        """Generate the HTML fragments."""
        results: List[str] = []
        for fragment in fragments:
            results.append("<!--StartFragment-->")
            results.append(f"{fragment}")
            results.append("<!--EndFragment-->")

        # Clean
        result: str = "\n".join(results)

        return result

    @staticmethod
    def _generate_html(string: str) -> str:
        """Generate the HTML document."""
        lines = string.splitlines()
        body = ["<body>"] + lines + ["</body>"]
        html = ["<html>"] + body + ["</html>"]

        return "\n".join(html)

    def _generate_header(self, string: str) -> str:
        """Generate the header for the HTML document."""
        lines = string.splitlines()

        version = self.version
        start_html_byte = self.start_html
        end_html_byte = self.end_html
        start_fragment_byte = self.start_fragment
        end_fragment_byte = self.end_fragment
        source_url = None

        if source_url is not None:
            lines.insert(0, f"SourceURL:{source_url}")
        lines.insert(0, f"EndFragment:{end_fragment_byte:0>10}")
        lines.insert(0, f"StartFragment:{start_fragment_byte:0>10}")
        lines.insert(0, f"EndHTML:{end_html_byte:0>10}")
        lines.insert(0, f"StartHTML:{start_html_byte:0>10}")
        lines.insert(0, f"Version:{version}")

        return "\n".join(lines)

    def _update_byte_counts(self, content: str) -> str:
        """Add byte counts to the HTML content."""
        # Setup
        content_bytes: bytes = content.encode(encoding=HTML_ENCODING)

        # Blocks to find
        html_start: bytes = "<html>".encode(encoding=HTML_ENCODING)
        html_end: bytes = "</html>".encode(encoding=HTML_ENCODING)
        fragment_start: bytes = "<!--StartFragment-->".encode(
            encoding=HTML_ENCODING
        )
        fragment_end: bytes = "<!--EndFragment-->".encode(
            encoding=HTML_ENCODING
        )

        # Find Values
        found_html_start: int = content_bytes.find(html_start)
        found_html_end: int = content_bytes.rfind(html_end) + len(html_end)
        found_fragment_start: int = (
            content_bytes.find(fragment_start) + len(fragment_start) + 1
        )  # after comment
        found_fragment_end: int = (
            content_bytes.rfind(fragment_end) - 1
        )  # before comment

        # Set Values
        self.start_html = found_html_start
        self.end_html = found_html_end
        self.start_fragment = found_fragment_start
        self.end_fragment = found_fragment_end

        # Update Values in HTML, with left-aligned 0s
        found_header_fragment_start: int = content_bytes.find(
            "StartFragment:".encode(HTML_ENCODING)
        )
        found_header_fragment_end: int = content_bytes.find(
            "EndFragment:".encode(HTML_ENCODING)
        )
        found_header_html_start: int = content_bytes.find(
            "StartHTML:".encode(HTML_ENCODING)
        )
        found_header_html_end: int = content_bytes.find(
            "EndHTML:".encode(HTML_ENCODING)
        )

        ## We need to update the integer values in the HTML content
        content_bytes = (
            # Up until the key
            content_bytes[:found_header_html_start]
            # The key and value
            + f"StartHTML:{self.start_html:0>10}".encode(HTML_ENCODING)
            # After the key and value
            + content_bytes[found_header_html_start + len("StartHTML:") + 10 :]
        )
        content_bytes = (
            # Up until the key
            content_bytes[:found_header_html_end]
            # The key and value
            + f"EndHTML:{self.end_html:0>10}".encode(HTML_ENCODING)
            # After the key and value
            + content_bytes[found_header_html_end + len("EndHTML:") + 10 :]
        )
        content_bytes = (
            # Up until the key
            content_bytes[:found_header_fragment_start]
            # The key and value
            + f"StartFragment:{self.start_fragment:0>10}".encode(HTML_ENCODING)
            # After the key and value
            + content_bytes[
                found_header_fragment_start + len("StartFragment:") + 10 :
            ]
        )
        content_bytes = (
            # Up until the key
            content_bytes[:found_header_fragment_end]
            # The key and value
            + f"EndFragment:{self.end_fragment:0>10}".encode(HTML_ENCODING)
            # After the key and value
            + content_bytes[
                found_header_fragment_end + len("EndFragment:") + 10 :
            ]
        )

        # Clean Up
        result = content_bytes.decode(encoding=HTML_ENCODING)

        return result



"""Clipboard module.

TODO: Make `get_clipboard`'s default to check available formats instead of just
using the default format.
FIXME: HTML_Format doesn't work for some reason. Fix this.
"""

import ctypes
import logging
import os
import time
import traceback
from typing import List
from typing import Optional
from typing import Union

from clipboard._c_interface import HANDLE
from clipboard._c_interface import LPVOID
from clipboard._c_interface import CloseClipboard
from clipboard._c_interface import EmptyClipboard
from clipboard._c_interface import EnumClipboardFormats
from clipboard._c_interface import GetClipboardData
from clipboard._c_interface import GlobalAlloc
from clipboard._c_interface import GlobalLock
from clipboard._c_interface import GlobalSize
from clipboard._c_interface import GlobalUnlock
from clipboard._c_interface import OpenClipboard
from clipboard._c_interface import SetClipboardData
from clipboard.constants import HTML_ENCODING
from clipboard.constants import UTF_ENCODING
from clipboard.errors import EmptyClipboardError
from clipboard.errors import FormatNotSupportedError
from clipboard.errors import GetClipboardError
from clipboard.errors import GetFormatsError
from clipboard.errors import LockError
from clipboard.errors import OpenClipboardError
from clipboard.errors import SetClipboardError
from clipboard.formats import ClipboardFormat
from clipboard.html_clipboard import HTMLTemplate


ClipboardFormatType = Union[int, str, ClipboardFormat]  # Type Alias
hMem = HANDLE  # Type Alias
GMEM_MOVEABLE = 0x0002
GMEM_ZEROINIT = 0x0040
GMEM_DDESHARE = 0x2000


logger = logging.getLogger(__name__)
formatter = logging.Formatter(
    "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
)
stream_handler = logging.StreamHandler()
stream_handler.setFormatter(formatter)
logger.addHandler(stream_handler)
if os.environ.get("LOGLEVEL"):
    logger.setLevel(os.environ["LOGLEVEL"])

    file_handler = logging.FileHandler("clipboard.log")
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)


def get_clipboard(
    format: Optional[ClipboardFormatType] = None,
) -> Optional[Union[str, bytes]]:
    """Convenience wrapper to get clipboard.

    Instead of using the `Clipboard.default_format`, this function uses the
    first available format on the clipboard.
    """
    if format is None:
        available = get_available_formats()
        if available:
            format = available[0]
    with Clipboard() as cb:
        return cb.get_clipboard(format=format)
    return None


def set_clipboard(
    content: Union[str, bytes],
    format: Optional[ClipboardFormatType] = None,
) -> HANDLE:
    """Convenience wrapper to set clipboard.

    Raises
    ------
    SetClipboardError
        If setting the clipboard failed.
    """
    with Clipboard() as cb:
        return cb.set_clipboard(content=content, format=format)
    raise SetClipboardError("Setting the clipboard failed.")


def get_available_formats() -> list[int]:
    """Convenience wrapper to get available formats.

    Raises
    ------
    OpenClipboardError
        If the context manager caught an error.
    """
    with Clipboard() as cb:
        return cb.available_formats()
    raise GetFormatsError("Failed to get available formats.")


class Clipboard:
    """Represents the system clipboard."""

    default_format: ClipboardFormat = ClipboardFormat.CF_UNICODETEXT

    def __init__(
        self,
        format: Optional[ClipboardFormatType] = None,
    ):
        if format is None:
            format = self.default_format.value
        else:
            format = self._resolve_format(format)

        self.format: int = format  # type: ignore
        self.h_clip_mem: Optional[HANDLE] = None
        self.address: Optional[LPVOID] = None
        self.size: Optional[int] = None  # in bytes

        self.locked: bool = False
        self.opened: bool = False

    def available_formats(self) -> List[int]:
        """Return all available clipboard formats on clipboard.

        First format is the format on the clipboard, depending on your system.
        """
        logger.info("Getting available clipboard formats")

        def get_formats() -> List[int]:
            formats: list[int] = [EnumClipboardFormats(0)]
            while formats[-1] != 0:
                formats.append(EnumClipboardFormats(formats[-1]))
            return formats[:-1]

        available_formats: List[int] = []
        if not self.opened:
            with self:
                available_formats = self.available_formats()
        elif self.opened:
            available_formats = get_formats()

        return available_formats

    def get_clipboard(
        self,
        format: Optional[ClipboardFormatType] = None,
    ) -> Optional[Union[str, bytes]]:
        """Get data from clipboard, returning None if nothing is on it.

        **Requires valid HTML content to work as intended.**

        Parameters
        ----------
        format : Optional[ClipboardFormatType]
            The format of the clipboard data.
            If None, the default format is used.

        Raises
        ------
        FormatNotSupportedError
            If the format is not supported.
        GetClipboardError
            If getting the clipboard data failed.
        LockError
            If locking the clipboard failed.
            If unlocking the clipboard failed.
        """
        logger.info("Getting clipboard data")

        if not self.opened:
            with self:
                return self.get_clipboard(format=format)

        if format is None:
            format = self.format
        else:
            format = self._resolve_format(format)

        formats = self.available_formats()
        if format not in formats:
            raise FormatNotSupportedError(
                f"{format} is not supported for getting the clipboard."
                f" Choose from following {formats}"
            )

        # Info
        self.h_clip_mem = GetClipboardData(format)
        if self.h_clip_mem is None:
            raise GetClipboardError("The `GetClipboardData` function failed.")
        self.address = self._lock(self.h_clip_mem)  # type: ignore
        self.size = GlobalSize(self.address)
        if not self.size:
            # 0 means that the function failed.
            raise GetClipboardError("The `GlobalSize` function failed.")

        # TODO: CF_LOCALE is special and should be handled differently.
        #   A handle to the locale identifier associated with the text in the
        #   clipboard.
        # TODO: There are other types that could be supported as well, such as
        # audio data:
        # https://learn.microsoft.com/en-us/windows/win32/dataxchg/standard-clipboard-formats
        string: ctypes.Array[ctypes.c_byte]
        content: Union[str, bytes]
        if format == ClipboardFormat.CF_UNICODETEXT.value:
            string = (ctypes.c_byte * self.size).from_address(
                int(self.address)  # type: ignore
            )
            content = bytearray(string)[:-2].decode(encoding=UTF_ENCODING)
        elif (
            format == ClipboardFormat.CF_HTML.value
            or format == ClipboardFormat.HTML_Format.value
        ):
            bytes_ = (ctypes.c_char * self.size).from_address(
                int(self.address)  # type: ignore
            )
            content = bytes(bytes_).decode(HTML_ENCODING)[:-1]

        else:
            string = (ctypes.c_byte * self.size).from_address(
                int(self.address)  # type: ignore
            )
            try:
                content = bytearray(string)[:-1].decode(encoding="utf-8")
            except UnicodeDecodeError:
                return bytes(string)

        # FIXME: This fails frequently, likely due to a resource management
        # error.
        try:
            self._unlock()
        except LockError:
            pass
        return content

    def set_clipboard(
        self,
        content: Union[str, bytes],
        format: Optional[ClipboardFormatType] = None,
    ) -> HANDLE:
        """Set clipboard.

        Raises
        ------
        SetClipboardError
            If setting the clipboard data failed.
        """
        logger.info("Setting clipboard data")

        set_handle: HANDLE = self._set_clipboard(content, format)

        return set_handle

    def _set_clipboard(
        self,
        content: Union[str, bytes],
        format: Optional[ClipboardFormatType] = None,
    ) -> HANDLE:
        """Hides the HANDLE.

        Raises
        ------
        SetClipboardError
            If setting the clipboard data failed.
        OpenClipboardError
            If opening the clipboard failed.
            Can be raised if the clipboard isn't already opened.
        EmptyClipboardError
            If emptying the clipboard failed.
            Need to empty the clipboard each time before setting new data.
        FormatNotSupportedError
            If the format is not supported.
        """

        logger.info("_Setting clipboard data")

        if format is None:
            format = self.format

        if not self.opened:
            with self:
                return self._set_clipboard(content=content, format=format)

        format = self._resolve_format(format)
        self._empty()

        set_handle: HANDLE
        alloc_handle: HANDLE
        content_bytes: bytes
        contents_ptr: LPVOID
        if format == ClipboardFormat.CF_UNICODETEXT.value:
            if isinstance(content, str):
                content_bytes = content.encode(encoding=UTF_ENCODING)
            else:
                content_bytes = content

            alloc_handle = GlobalAlloc(
                GMEM_MOVEABLE | GMEM_ZEROINIT, len(content_bytes) + 2
            )
            contents_ptr = GlobalLock(alloc_handle)
            ctypes.memmove(contents_ptr, content_bytes, len(content_bytes))
            GlobalUnlock(alloc_handle)

            set_handle = SetClipboardData(format, alloc_handle)

        elif (
            format == ClipboardFormat.CF_HTML.value
            or format == ClipboardFormat.HTML_Format.value
        ):
            content_str: str  # utf-8
            if isinstance(content, bytes):
                content_str = content.decode(encoding=HTML_ENCODING)
            else:
                content_str = content  # type: ignore
            template: HTMLTemplate = HTMLTemplate(content_str)
            html_content_bytes: bytes = template.generate().encode(
                encoding=HTML_ENCODING
            )

            alloc_handle = GlobalAlloc(
                GMEM_MOVEABLE | GMEM_ZEROINIT, len(html_content_bytes) + 1
            )
            contents_ptr = GlobalLock(alloc_handle)  # type: ignore
            ctypes.memmove(
                contents_ptr, html_content_bytes, len(html_content_bytes)
            )
            GlobalUnlock(alloc_handle)

            set_handle = SetClipboardData(format, alloc_handle)
        else:
            if isinstance(content, str):
                # Most general content is going to be utf-8.
                content_bytes = content.encode(encoding="utf-8")
            else:
                content_bytes = content

            alloc_handle = GlobalAlloc(GMEM_MOVEABLE, len(content_bytes) + 1)
            contents_ptr = GlobalLock(alloc_handle)
            ctypes.memmove(contents_ptr, content_bytes, len(content_bytes))
            GlobalUnlock(alloc_handle)

            set_handle = SetClipboardData(format, alloc_handle)

        if set_handle is None:
            raise SetClipboardError("Setting the clipboard failed.")

        return set_handle

    def _resolve_format(self, format: ClipboardFormatType) -> int:
        """Given an integer, representing a clipboard format, or a
        ClipboardFormat object, return the respective integer.

        Raises
        ------
        FormatNotSupportedError
            If the format is not supported.
        """

        logger.info("Resolving clipboard format")

        if isinstance(format, ClipboardFormat):
            format = format.value
        elif isinstance(format, int):
            pass
        elif isinstance(format, str):
            try:
                format = ClipboardFormat[format].value
            except KeyError as exc:
                formats = self.available_formats()
                raise FormatNotSupportedError(
                    f"{format} is not a supported clipboard format."
                    f" Choose from following {formats}"
                ) from exc

        # FIXME: There are issues with HTML_Format, so use CF_HTML
        if format == ClipboardFormat.HTML_Format.value:
            format = ClipboardFormat.CF_HTML.value
        return format  # type: ignore

    def __getitem__(self, format: ClipboardFormatType):
        """Get data from clipboard, returning None if nothing is on it.

        Raises
        ------
        FormatNotSupportedError
            If the format is not supported.
        GetClipboardError
            If getting the clipboard data failed.
        """
        return self.get_clipboard(format)

    def __setitem__(self, format, content) -> None:
        """Set clipboard content based on given format.

        Raises
        ------
        SetClipboardError
            If setting the clipboard data failed.
        FormatNotSupportedError
            If the given format is not supported.
        OpenClipboardError
            Can only be raised if the clipboard isn't already opened.
        EmptyClipboardError
            If emptying the clipboard failed.
            The clipboard needs to be emptied before setting new data.
        """
        format = self._resolve_format(format)
        self._empty()
        self.set_clipboard(content, format)

    def __enter__(self):
        """Open clipboard.

        Raises
        ------
        OpenClipboardError
            If opening the clipboard failed.
            Can only be raised if the clipboard isn't already opened.
        """
        logger.info("Entering context manager")

        # There is an issue with opening the clipboard repeatedly.
        # This is likely due to a logical error when locking and unlocking.
        #
        # This successfully fixes the issue, although it is blunt.
        max_tries = 3
        tries = 0
        while not self.opened and tries < max_tries:
            if tries > 0:
                time.sleep(0.01)
            tries += 1
            if self._open():
                return self
            self._close()

        raise OpenClipboardError("Failed to open clipboard.")

    def __exit__(
        self, exception_type, exception_value, exception_traceback
    ) -> bool:
        logger.info("Exiting context manager")
        if exception_type is not None:
            traceback.print_exception(
                exception_type, exception_value, exception_traceback
            )

        self._close()
        return True

    def _open(self, handle: Optional[HANDLE] = None) -> bool:
        logger.info("_Opening clipboard")
        opened: bool = bool(OpenClipboard(handle))
        self.opened = opened
        return opened

    def _close(self) -> bool:
        logger.info("_Closing clipboard")
        self.opened = False
        # FIXME: This fails frequently, likely due to a resource management
        # error.
        try:
            self._unlock()
        except LockError:
            pass
        return CloseClipboard()

    def _lock(self, handle: HANDLE) -> LPVOID:
        """Lock clipboard.

        Raises
        ------
        LockError
            If locking the clipboard failed.
        """
        logger.info("_Locking clipboard")
        locked: LPVOID = GlobalLock(handle)
        self.locked = bool(locked)
        if locked is None:
            raise LockError("The `GlobalLock` function failed.")

        return locked

    def _unlock(self, handle: Optional[HANDLE] = None) -> bool:
        """Unlock clipboard.

        Raises
        ------
        LockError
            If unlocking the clipboard failed.
        """
        logger.info("_Unlocking clipboard")
        if handle is None:
            handle = self.h_clip_mem

        if handle is None:
            # Clipboard is already unlocked
            return True

        if not self.locked:
            # Clipboard is already unlocked
            return True

        unlocked: bool = bool(GlobalUnlock(handle))  # non-zero
        self.locked = not unlocked
        if not unlocked:
            raise LockError("The `GlobalUnlock` function failed.")
        return unlocked

    def _empty(self) -> bool:
        """Empty clipboard.

        Raises
        ------
        EmptyClipboardError
            If emptying the clipboard failed.
        """
        logger.info("_Emptying clipboard")
        if not self.opened:
            with self:
                return self._empty()
        elif self.opened:
            # FIXME: A false means that this failed.
            return bool(EmptyClipboard())
        else:
            raise EmptyClipboardError("Emptying the clipboard failed.")

        return False



"""Clipboard errors."""


class FormatNotSupportedError(Exception):
    """Exception raised when a clipboard format is not supported."""


class GetClipboardError(Exception):
    """Exception raised when getting the clipboard fails."""


class SetClipboardError(Exception):
    """Exception raised when setting the clipboard fails."""


class EmptyClipboardError(Exception):
    """Exception raised when emptying the clipboard fails."""


class OpenClipboardError(Exception):
    """Exception raised when opening the clipboard fails."""


class ClipboardError(Exception):
    """Exception raised when clipboard operation fails."""


class LockError(Exception):
    """Exception raised when locking the clipboard fails."""


class GetFormatsError(Exception):
    """Exception raised when getting clipboard formats fails."""
